//-----------------------------------------------------------------------------
// name: VisualPam.cpp
// desc: Psycho-acoustic Model Visualizer
//
// Stanford University | Music 256a: Music, Computing, and Design
//     http://ccrma.stanford.edu/courses/256a/
//-----------------------------------------------------------------------------
#include "RtAudio.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "chuck_fft.h"
#include "keun_pam.h"

#ifdef __MACOSX_CORE__
  // note: for mac only
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <iostream>
using namespace std;


// some defines
#define SAMPLE double
#define MY_PIE 3.14159265358979
#define MY_SRATE 44100

#define	X_AXIS_LOG	0
#define	LOG_AXIS_SCALE	0.1

#define	INC_STEP	0.025f

#define	BANDWIDTH	.5


// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;

// globle buffer
SAMPLE ** g_buffer;
float * g_specbuff;
float * g_timebuff;
float * g_histbuff;
float * g_window;
float * g_timewindow;
float * g_msr[2];
float * g_thr;
float * g_en;
float g_pe[2];
long g_blkType[2];
long *g_blockSwBuff;

// global variables
long g_bufferSize = 2048; // size of buffer in samples
unsigned long g_windowSize = g_bufferSize << 1;
unsigned long g_FFTsize = g_windowSize >> 1;
long g_layerDepth = 48;
long g_nframe = 0;
int g_buffwptr = 0;
int g_windowType = 0;
int g_frameSkip = 1;
bool g_ready = false;
bool g_waterfall = true;
bool g_logAxis = false;
bool g_freeze = false;
bool g_psycho = true;

GLfloat g_view_z = 0.;
GLfloat g_rotate_y = 0.;
GLfloat g_bandWidth = 0.5;
GLfloat g_distance = 7.7f;
GLfloat g_freqGain = 50.f;
GLfloat g_timeGain = 5.f;

KeunPAM *pam;

//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void specialFunc( int key, int x, int y );
void helpScreen( );
void draw_string( GLfloat x, GLfloat y, GLfloat z, const char * str, GLfloat scale );




// audio callback
int callme( void * out, void * in, unsigned int numFrames, double streamTime,
			RtAudioStreamStatus status, void * user_data )
{
	if( status )
		std::cout << "Stream over/underflow detected." << std::endl;

    SAMPLE * buffin = (SAMPLE *)in;
    SAMPLE * buffy = (SAMPLE *)out;

	memcpy( g_histbuff, g_histbuff+numFrames, sizeof(float)*numFrames);

	for( int n=0; n < numFrames; n++ ){
		g_histbuff[n+numFrames] = (float)buffin[n];
		buffy[n] = buffin[n];
	}

	if( ++g_nframe%g_frameSkip == 0 )
		g_ready = true;

    return 0;
}




// entry point
int main( int argc, char ** argv )
{
    // initialize GLUT
    glutInit( &argc, argv );
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "VisualPAM" );
    
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
	// set the special function - called on special key events
	glutSpecialFunc( specialFunc );

	// help messege
	helpScreen();

	// PAM
	pam = new KeunPAM(MY_SRATE);
 
    // RtAudio pointer
    RtAudio audio;

    // buffer size
    unsigned int buffer_size = g_bufferSize;

	// parameters
	RtAudio::StreamParameters iParams, oParams;
	iParams.deviceId = audio.getDefaultInputDevice();
	iParams.nChannels = 1;
	oParams.deviceId = audio.getDefaultOutputDevice();
	oParams.nChannels = 1;
    
    // create the RtAudio
    try {
        audio.openStream(
			&oParams,
			&iParams,
            RTAUDIO_FLOAT64, // format
            MY_SRATE, // sample rate
            &buffer_size, // buffer size
			callme,	// callback function
			NULL
        );
    } catch( RtError & err ) {
        err.printMessage();
        exit(1);
    }
    
    // allocate global buffer
	g_specbuff = new float[g_windowSize];
	g_histbuff = new float[g_windowSize];
	g_timebuff = new float[buffer_size];
	g_window = new float[g_windowSize];
	g_timewindow = new float[g_windowSize];

	g_buffer = new SAMPLE *[g_layerDepth];
	for( int n=0; n < g_layerDepth; n++ ){
		g_buffer[n] = new SAMPLE[buffer_size];
		memset(g_buffer[n], 0, sizeof(SAMPLE)*buffer_size);
	}

	g_en = new float[pam->getNumSubBand()];
	memset(g_en, 0, sizeof(float)*pam->getNumSubBand());

	g_thr = new float[pam->getNumSubBand()];
	memset(g_thr, 0, sizeof(float)*pam->getNumSubBand());

	g_msr[0] = new float[buffer_size>>1];

	g_blockSwBuff = new long[g_layerDepth];

	memset( g_histbuff, 0, sizeof(float)*g_windowSize);

	// making window
	hanning(g_window, g_windowSize);
	hanning(g_timewindow, buffer_size);
	pam->setFFTWinType(HANNING_WINDOW);

    // set the callback
    try {
        audio.startStream();
    } catch( RtError & err ) {
        // do stuff
        err.printMessage();
        goto cleanup;
    }

    // let GLUT handle the current thread from here
    glutMainLoop();

cleanup:
	if( audio.isStreamOpen() )
	    audio.closeStream();

	delete[] g_specbuff;
	delete[] g_histbuff;
	delete[] g_timebuff;
	delete[] g_window;
	delete[] g_timewindow;

	for( int n=0; n < g_layerDepth; n++ )
		delete []g_buffer[n];
	delete[] g_buffer;

	delete[] g_en;
	delete[] g_thr;
	delete[] g_msr[0];
   
	delete[] g_blockSwBuff; 

    return 0;
}



//-----------------------------------------------------------------------------
// Name: helpScreen( )
// Desc: ...
//-----------------------------------------------------------------------------
void helpScreen( )
{
	std::cerr << "\n=======================================================" << std::endl;
	std::cerr << " Psycho-acoustic Model Visualizer (v0.5)" << std::endl;
	std::cerr << "=======================================================" << std::endl;
	std::cerr << " Keun Sup Lee, CCRMA, Stanford University" << std::endl;
	std::cerr << "-------------------------------------------------------" << std::endl;
	std::cerr << "[Usage]" << std::endl;
	std::cerr << " 'p' : toggle Psycho-acoustic model info display" << std::endl;
	std::cerr << " 'n' : use haNning window for computing spectrum" << std::endl;
	std::cerr << " 'm' : use haMming window for computing spectrum" << std::endl;
	std::cerr << " 'k' : use blacKman window for computing spectrum" << std::endl;
	std::cerr << " 'c' : use reCtanglar window for computing spectrum" << std::endl;
	std::cerr << " '1' : spectrum gain boost" << std::endl;
	std::cerr << " '2' : spectrum gain cut" << std::endl;
	std::cerr << " 'q' : Quit this program" << std::endl;

	std::cerr << "\n - display setting -" << std::endl;
	std::cerr << " 'r' : Reset view point" << std::endl;
	std::cerr << " 'f' : Freezing display" << std::endl;
	std::cerr << " 'w' : toggle spectrum display (default: Waterfall on)" << std::endl;
	std::cerr << " 'l' : toggle Log axis (default: off)" << std::endl;
	std::cerr << " ']' : increase spectrum bandwidth" << std::endl;
	std::cerr << " '[' : decrease spectrum bandwidth" << std::endl;
	std::cerr << " '+' : increase display frame per second" << std::endl;
	std::cerr << " '-' : decrease display frame per second" << std::endl;
	std::cerr << " [left arrow key] : move around to the Left" << std::endl;
	std::cerr << " [right arrow key] : move around to the Right" << std::endl;
	std::cerr << " [up arrow key] : move Up" << std::endl;
	std::cerr << " [down arrow key] : move Down" << std::endl;
	std::cerr << " [page up key] : going closer" << std::endl;
	std::cerr << " [page down key] : going away\n" << std::endl;


}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, g_distance*sin(g_view_z), g_distance*cos(g_view_z), 0.0f, 0.0f, 0.0f, 0.0f, (cos(g_view_z) < 0 ? -1.f : 1.0f), 0.0f );
}



//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;

		case 'R':
		case 'r':
			g_view_z = 0.f;
			g_rotate_y = 0.f;
			g_bandWidth = 0.5f;
			g_distance = 7.7f;
			g_frameSkip = 1;
			g_freqGain = 50.f;
			g_timeGain = 5.f;
			break;

		case 'W':
		case 'w':
			g_waterfall = !g_waterfall;
			break;

		case 'L':
		case 'l':
			g_logAxis = !g_logAxis;
			break;

		case 'F':
		case 'f':
			g_freeze = !g_freeze;
			break;

		case '-':
			if( ++g_frameSkip > 5 )
				g_frameSkip = 5;
			break;

		case '+':
			if( --g_frameSkip < 1 )
				g_frameSkip = 1;
			break;

		case 'N':
		case 'n':
			hanning(g_window, g_windowSize);
			hanning(g_timewindow, g_windowSize>>1);
			g_windowType = HANNING_WINDOW;
			pam->setFFTWinType(HANNING_WINDOW);
			break;

		case 'M':
		case 'm':
			hamming(g_window, g_windowSize);
			hamming(g_timewindow, g_windowSize>>1);
			g_windowType = HAMMING_WINDOW;
			pam->setFFTWinType(HAMMING_WINDOW);
			break;

		case 'K':
		case 'k':
			blackman(g_window, g_windowSize);
			blackman(g_timewindow, g_windowSize>>1);
			g_windowType = BLACKMAN_WINDOW;
			pam->setFFTWinType(BLACKMAN_WINDOW);
			break;

		case 'C':
		case 'c':
			for( int n=0; n < g_windowSize; n++ )
				g_window[n] = 1.f;

			for( int n=0; n < g_windowSize>>1; n++ )
				g_timewindow[n] = 1.f;

			g_windowType = RECT_WINDOW;
			pam->setFFTWinType(RECT_WINDOW);
			break;

		case '[':
			g_bandWidth -= 2000.f/MY_SRATE;
			if( (int)(MY_SRATE/2*g_bandWidth) < 4000 )
				g_bandWidth = 8000.f/MY_SRATE;
			break;

		case ']':
			g_bandWidth += 2000.f/MY_SRATE;
			if( g_bandWidth > 1.f )
				g_bandWidth = 1.f;
			break;

		case '1':
			g_freqGain *= pow(10., 1./20.);
			g_timeGain *= pow(10., 1./20.);
			break;

		case '2':
			g_freqGain /= pow(10., 1./20.);
			g_timeGain /= pow(10., 1./20.);
			break;

		case 'p':
			g_psycho = !g_psycho;
			break;
    }
    
	reshapeFunc( g_width, g_height );
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else
    {
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles special keys
//-----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
	if( key == GLUT_KEY_LEFT )
		g_rotate_y += INC_STEP*50;
	else if( key == GLUT_KEY_RIGHT )
		g_rotate_y -= INC_STEP*50;
	else if( key == GLUT_KEY_UP )
		g_view_z += INC_STEP;
	else if( key == GLUT_KEY_DOWN )
		g_view_z -= INC_STEP;
	else if( key == GLUT_KEY_PAGE_UP )
		g_distance -= INC_STEP*2;
	else if( key == GLUT_KEY_PAGE_DOWN )
		g_distance += INC_STEP*2;

	reshapeFunc( g_width, g_height );
	glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}



//-----------------------------------------------------------------------------
// Name: draw_string()
// Desc: ...
//-----------------------------------------------------------------------------
void draw_string( GLfloat x, GLfloat y, GLfloat z, const char * str, GLfloat scale = 1.0f )
{
    GLint len = strlen( str ), i;

    glPushMatrix();
    glTranslatef( x, y, z );
    glScalef( .001f * scale, .001f * scale, .001f * scale );

    for( i = 0; i < len; i++ )
        glutStrokeCharacter( GLUT_STROKE_ROMAN, str[i] );

    glPopMatrix();
}



//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
	long count=0;
	static int timeBase=0;

    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // line width
    glLineWidth( 1.0 );
    // step through and plot the waveform
	GLfloat xstart = -1.5;
    GLfloat x = -1.5;

    // increment
    GLfloat xinc = ::fabs(2*x / (g_bufferSize*g_bandWidth));
	GLfloat xscale = ::fabs(2*x) / log10(g_bufferSize*g_bandWidth);

	static GLfloat pe=0;

	// wait for reading
	while( !g_ready ){
		usleep( 1000 );

		// escape condition for preventing deadlock
		if( ++count > 100 ) break;
	}


	if( !g_freeze ) {

		//////////////////////////////////////////////////////////////////
		// calculating spectrum

		// Psycho Acoustic Model
		pam->PsychoAcousticModel(g_histbuff, g_pe, g_msr, g_blkType, 1);

		memcpy(g_specbuff, g_histbuff, sizeof(float)*g_windowSize); 

		// applying window
		apply_window(g_specbuff, g_window, g_windowSize);
	
		// Forward FFT
		rfft(g_specbuff, g_FFTsize, FFT_FORWARD);
	
		complex * pCmp = (complex *)g_specbuff;

		// copy to global buffer
		for( int n = 0; n < g_bufferSize>>1; n++ )
			g_buffer[g_buffwptr][n] = (SAMPLE)20.*log10(cmp_abs(pCmp[n])*g_freqGain+1);

		// copy to block type display buffer
		g_blockSwBuff[g_buffwptr] = (g_blkType[0] == SHORT_WINDOW ? 1 : 0);


		float *p_thr = g_thr;
		float *p_en = g_en;
		long maxSB = pam->getNumSubBand();
		long *CBwidthTab = pam->getCBwidthTab();

		int j=0;
		// calculate masking threshold
		for( int n=0; n < maxSB; n++ ){
			float sumThr= 0.f;
			float sumEn= 0.f;
			for( int m=0; m < CBwidthTab[n]; m++, j++ ){
				sumThr += 20.f*log10(cmp_abs(pCmp[j])*g_freqGain*g_msr[0][j>>1]+1);
				sumEn += 20.f*log10(cmp_abs(pCmp[j])*g_freqGain+1);
			}
			*p_thr++ = (*p_thr+sumThr)/2.f;	// slew control
			*p_en++ = (*p_en+sumEn)/2.f;
		}

		// calculate perceptual entropy
		pe = .5f*pe + .5f*g_pe[0];			// slew control
	}

    // push the matrix
	glPushMatrix();

		glRotatef( g_rotate_y, 0.f, 1.f, 0.f);

		glTranslatef( 0.f, 0.f, 3.0f);


		////////////////////////////////////////////////
		// displaying frequency domain spectrum

		// push the matrix
		glPushMatrix();

		// position the spectrum
		glTranslatef( 0.f, -1.5f, 0.f);

		glScalef( 1.5f, 1.f, 1.f);	

		// set the color
		glColor3f( .5f, 1.f, .5f );

		if( g_waterfall ) {

			for( int n=g_layerDepth; n > 0; n-- ){
				int layerIndex = (g_buffwptr-n)%g_layerDepth;
				double colorScale = (double)(g_layerDepth-n)/g_layerDepth;

				if( layerIndex < 0 )
					layerIndex += g_layerDepth;

   			 	// go
			    glBegin( GL_LINE_STRIP );

				// set the color
				glColor3f( .5f*colorScale, 1.f*colorScale, .5f*colorScale );

			    // loop through global buffer
				if( g_logAxis ) {
	   				for( int i = 0; i < (g_bufferSize>>1)*g_bandWidth; i++ )
   					{
						// x-position
						x = xstart + log10(2*i+1)*xscale;

						// set the next vertex
						glVertex3f( x, g_buffer[layerIndex][i], -(n-1)*0.2 );
					}
				} else {
   					for( int i = 0; i < (g_bufferSize>>1)*g_bandWidth; i++ )
   					{
						// x-position
						x = xstart + xinc*2*i;

						// set the next vertex
						glVertex3f( x, g_buffer[layerIndex][i], -(n-1)*0.2 );
					}
				}
   			 	// done
   			 	glEnd();
			}
		} else {

	  		// go
			glBegin( GL_LINE_STRIP );

			// loop through global buffer
			if( g_logAxis ) {
				for( int i = 0; i < g_bufferSize*g_bandWidth; i++ )
  				{
					// x-position
					x = xstart + log10(i+1)*xscale;

					// set the next vertex
					glVertex2f( x, g_buffer[g_buffwptr][i] );
				}
			} else {
				for( int i = 0; i < g_bufferSize*g_bandWidth; i++ )
  				{
					// x-position
					x = xstart + xinc*i;

					// set the next vertex
					glVertex2f( x, g_buffer[g_buffwptr][i] );
				}
			}

   			 // done
   			 glEnd();

		}

		// pop
		glPopMatrix();



		if( g_psycho ) {
			////////////////////////////////////////////////
			// displaying block type info

			// push the matrix
			glPushMatrix();

			// position the spectrum
			glTranslatef( 0.f, -1.5f, 0.f);

			glScalef( 1.5f, 1.f, 1.f);	

			// set the color
			glColor3f( .5f, .5f, 1.f );

			if( g_waterfall ) {

				// go
				glBegin( GL_LINE_STRIP );

				for( int n=g_layerDepth; n > 0; n-- ){
					int layerIndex = (g_buffwptr-n)%g_layerDepth;
					double colorScale = (double)(g_layerDepth-n)/g_layerDepth;

					if( layerIndex < 0 )
						layerIndex += g_layerDepth;

					// set the color
					glColor3f( .5f*colorScale, .5f*colorScale, 1.f*colorScale );

					// set the next vertex
					glVertex3f( xstart, 0.5f*g_blockSwBuff[layerIndex], -(n-1)*0.2 );
				}

				// done
   				glEnd();
			}

			// pop
			glPopMatrix();



			////////////////////////////////////////////////
			// displaying psycho-acoustic model

			// push the matrix
			glPushMatrix();

			// position the spectrum
			glTranslatef( .5f, .2f, 0.f);

			glScalef( 1.3f, 1.f, 1.f);	

  			// go
			glBegin( GL_LINE_STRIP );

			// set the color
			glColor3f( 1.f, .5f, .5f );

			x = xstart;
			GLfloat inc = ::fabs(1.2f*x / pam->getNumSubBand());

			for( int i = 0; i < pam->getNumSubBand(); i++ ){
				// set the next vertex
				glVertex2f( x, g_thr[i] );
				// x-position
				x += inc;
			}

   			// done
		   	glEnd();

  			// go
			glBegin( GL_LINE_STRIP );

			// set the color
			glColor3f( 1.f, 1.f, 1.f );

			x = xstart;

			for( int i = 0; i < pam->getNumSubBand(); i++ ){
				// set the next vertex
				glVertex2f( x, g_en[i] );
				// x-position
				x += inc;
			}

   			// done
		   	glEnd();

			// displaying perceptual entropy

			// push the matrix
			glPushMatrix();

				// line width
				glLineWidth( 20.0 );

  				// go
				glBegin( GL_LINE_STRIP );

				// set the color
				glColor3f( 1.f, 1.f, .5f );

				x = .4f;
				glVertex2f( x, 0 );
				glVertex2f( x, pe/1000.f );

   				// done
		   		glEnd();

				// line width
				glLineWidth( 1.0 );

			// pop
			glPopMatrix();

			// pop
			glPopMatrix();

		} else {	// if( g_psycho )

        	////////////////////////////////////////////////
        	// displaying time domain signal

        	// push the matrix
			glPushMatrix();

        	// position the spectrum
        	glTranslatef( 0.f, .7f, 0.f);

        	glScalef( 1.3f, 1.f, 1.f);

        	// go
        	glBegin( GL_LINE_STRIP );

        	// set the color
        	glColor3f( 1.f, .5f, .5f );

        	if( !g_freeze ) {
           		memcpy(g_timebuff, g_histbuff+g_bufferSize, sizeof(float)*g_bufferSize);
            	apply_window(g_timebuff, g_timewindow, g_bufferSize);
        	}

        	x = xstart;

        	// loop through global buffer
        	for( int i = 0; i < g_bufferSize; i++ ){
            	// set the next vertex
            	glVertex2f( x, g_timeGain*g_timebuff[i] );
            	// x-position
            	x += xinc*g_bandWidth;
        	}

        	// done
        	glEnd();

        	// pop
        	glPopMatrix();
		}


		////////////////////////////////////////////////
		// displaying text string

		glColor3f( 1.f, 1.f, 1.f );

		char string[100];
		sprintf(string, "BW: %d kHz", (int)(MY_SRATE/2*g_bandWidth/1000.f));
		draw_string( 1.8f, -1.6f, 0.f, string, .5f);

		glColor3f( 1.f, 1.f, 0.5f );

		// count frame per second
		int time = glutGet(GLUT_ELAPSED_TIME);
		int frameRate = 1000.0f/(time-timeBase);
		timeBase = time;

		if( g_freeze )
			frameRate = 0;

		// display Frame Per Second
		sprintf(string, "Estimated FPS: %2d fps", frameRate);
		draw_string( 1.5f, -1.8f, 0.f, string, .6f);

		int gain = g_freqGain >= 50.f ? (int)(20.*log10(g_freqGain/50.) + .5) :
										(int)(20.*log10(g_freqGain/50.) - .5);

		// display spectrum gain
		sprintf(string, "Gain boost: % d dB", gain);
		draw_string( -2.3f, -1.8f, 0.f, string, .6f);

		// display axis
		if( g_logAxis )
			sprintf(string, "X-Axis: Log scale", gain);
		else
			sprintf(string, "X-Axis: Linear scale", gain);
		draw_string( -1.3f, -1.8f, 0.f, string, .6f);

		switch( g_windowType ){
			case HANNING_WINDOW:
				sprintf(string, "Window: Hanning");
				break;
			case HAMMING_WINDOW:
				sprintf(string, "Window: Hamming");
				break;
			case BLACKMAN_WINDOW:
				sprintf(string, "Window: Blackman");
				break;
			case RECT_WINDOW:
				sprintf(string, "Window: Rectangular");
				break;
		}
		draw_string( -.3f, -1.8f, 0.f, string, .6f);

		// display perceptual entropy
		if( g_psycho ) {
			glColor3f( 1.f, 1.f, 1.f );
			sprintf(string, "PE: %6.2f", pe);
			draw_string( 1.1f, 0.2f, 0.f, string, .5f);

			sprintf(string, "critical band");
			draw_string( .45f, 0.1f, 0.f, string, .5f);
		}

		// increasing buffer read pointer
		if( !g_freeze )
			g_buffwptr = (g_buffwptr+1)%g_layerDepth;
		g_ready = false;

   	// pop
   	glPopMatrix();


    // flush!
    glFlush( );


    // swap the double buffer
    glutSwapBuffers( );
}
